@@ -15,6 +15,7 @@ use libc::c_char;
 
 use crate::arena::Handle;
 use crate::graph::File;
+use crate::graph::InternedString;
 use crate::graph::Node;
 use crate::graph::NodeID;
 use crate::graph::StackGraph;
@@ -34,6 +35,7 @@ use crate::paths::ScopeStack;
 use crate::paths::ScopedSymbol;
 use crate::paths::SymbolStack;
 use crate::stitching::Database;
+use crate::stitching::ForwardPartialPathStitcher;
 use crate::stitching::PathStitcher;
 
 /// Contains all of the nodes and edges that make up a stack graph.
@@ -120,6 +122,9 @@ pub extern "C" fn sg_partial_path_database_free(db: *mut sg_partial_path_databas
     drop(unsafe { Box::from_raw(db) })
 }
 
+/// The null value for all of our handles.
+pub const SG_NULL_HANDLE: u32 = 0;
+
 /// The handle of an empty list.
 pub const SG_LIST_EMPTY_HANDLE: u32 = 0xffffffff;
 
@@ -217,6 +222,81 @@ pub extern "C" fn sg_stack_graph_add_symbols(
     }
 }
 
+//-------------------------------------------------------------------------------------------------
+// Interned strings
+
+/// Arbitrary string content associated with some part of a stack graph.
+#[repr(C)]
+pub struct sg_string {
+    pub content: *const c_char,
+    pub length: usize,
+}
+
+/// A handle to an interned string in a stack graph.  A zero handle represents a missing string.
+///
+/// We deduplicate strings in a stack graph — that is, we ensure that there are never multiple
+/// `struct sg_string` instances with the same content.  That means that you can compare string
+/// handles using simple equality, without having to dereference them.
+pub type sg_string_handle = u32;
+
+/// An array of all of the interned strings in a stack graph.  String handles are indices into this
+/// array. There will never be a valid string at index 0; a handle with the value 0 represents a
+/// missing string.
+#[repr(C)]
+pub struct sg_strings {
+    pub strings: *const sg_string,
+    pub count: usize,
+}
+
+/// Returns a reference to the array of string data in this stack graph.  The resulting array
+/// pointer is only valid until the next call to any function that mutates the stack graph.
+#[no_mangle]
+pub extern "C" fn sg_stack_graph_strings(graph: *const sg_stack_graph) -> sg_strings {
+    let graph = unsafe { &(*graph).inner };
+    sg_strings {
+        strings: graph.strings.as_ptr() as *const sg_string,
+        count: graph.strings.len(),
+    }
+}
+
+/// Adds new strings to the stack graph.  You provide all of the string content concatenated
+/// together into a single string, and an array of the lengths of each string.  You also provide an
+/// output array, which must have the same size as `lengths`.  We will place each string's handle
+/// in the output array.
+///
+/// We ensure that there is only ever one copy of a particular string stored in the graph — we
+/// guarantee that identical strings will have the same handles, meaning that you can compare the
+/// handles using simple integer equality.
+///
+/// We copy the string data into the stack graph.  The string content you pass in does not need to
+/// outlive the call to this function.
+///
+/// Each string must be a valid UTF-8 string.  If any string isn't valid UTF-8, it won't be added
+/// to the stack graph, and the corresponding entry in the output array will be the null handle.
+#[no_mangle]
+pub extern "C" fn sg_stack_graph_add_strings(
+    graph: *mut sg_stack_graph,
+    count: usize,
+    strings: *const c_char,
+    lengths: *const usize,
+    handles_out: *mut sg_string_handle,
+) {
+    let graph = unsafe { &mut (*graph).inner };
+    let mut strings = strings as *const u8;
+    let lengths = unsafe { std::slice::from_raw_parts(lengths, count) };
+    let handles_out = unsafe {
+        std::slice::from_raw_parts_mut(handles_out as *mut Option<Handle<InternedString>>, count)
+    };
+    for i in 0..count {
+        let string = unsafe { std::slice::from_raw_parts(strings, lengths[i]) };
+        handles_out[i] = match std::str::from_utf8(string) {
+            Ok(string) => Some(graph.add_string(string)),
+            Err(_) => None,
+        };
+        unsafe { strings = strings.add(lengths[i]) };
+    }
+}
+
 //-------------------------------------------------------------------------------------------------
 // Files
 
@@ -436,11 +516,14 @@ pub extern "C" fn sg_stack_graph_nodes(graph: *const sg_stack_graph) -> sg_nodes
 /// You cannot add new instances of the root node or "jump to scope" node, since those are
 /// singletons and already exist in the stack graph.
 ///
+/// If you try to add a new node that has the same ID as an existing node in the stack graph, the
+/// new node will be ignored, and the corresponding entry in the `handles_out` array will contain
+/// the handle of the _existing_ node with that ID.
+///
 /// If any node that you pass in is invalid, it will not be added to the graph, and the
-/// corresponding entry in the `handles_out` array will be null.  (Note that includes trying to add
-/// a node with the same ID as an existing node, since all nodes must have unique IDs.)
+/// corresponding entry in the `handles_out` array will be null.
 #[no_mangle]
-pub extern "C" fn sg_stack_graph_add_nodes(
+pub extern "C" fn sg_stack_graph_get_or_create_nodes(
     graph: *mut sg_stack_graph,
     count: usize,
     nodes: *const sg_node,
@@ -452,8 +535,8 @@ pub extern "C" fn sg_stack_graph_add_nodes(
         unsafe { std::slice::from_raw_parts_mut(handles_out as *mut Option<Handle<Node>>, count) };
     for i in 0..count {
         let node_id = nodes[i].id;
-        handles_out[i] =
-            validate_node(graph, &nodes[i]).and_then(|node| graph.add_node(node_id.into(), node));
+        handles_out[i] = validate_node(graph, &nodes[i])
+            .map(|node| graph.get_or_create_node(node_id.into(), node));
     }
 }
 
@@ -534,6 +617,120 @@ pub extern "C" fn sg_stack_graph_add_edges(
     }
 }
 
+//-------------------------------------------------------------------------------------------------
+// Source info
+
+/// Contains information about a range of code in a source code file.
+#[repr(C)]
+#[derive(Clone, Copy, Default)]
+pub struct sg_source_info {
+    /// The location in its containing file of the source code that this node represents.
+    pub span: sg_span,
+    /// The kind of syntax entity this node represents (e.g. `function`, `class`, `method`, etc.).
+    pub syntax_type: sg_string_handle,
+    /// The full content of the line containing this node in its source file.
+    pub containing_line: sg_string_handle,
+}
+
+/// All of the position information that we have about a range of content in a source file
+#[repr(C)]
+#[derive(Clone, Copy, Default)]
+pub struct sg_span {
+    pub start: sg_position,
+    pub end: sg_position,
+}
+
+/// All of the position information that we have about a character in a source file
+#[repr(C)]
+#[derive(Clone, Copy, Default)]
+pub struct sg_position {
+    /// The 0-indexed line number containing the character
+    pub line: usize,
+    /// The offset of the character within its containing line, expressed as both a UTF-8 byte
+    /// index and a UTF-16 code point index
+    pub column: sg_offset,
+    /// The UTF-8 byte indexes (within the file) of the start and end of the line containing the
+    /// character
+    pub containing_line: sg_utf8_bounds,
+    /// The UTF-8 byte indexes (within the file) of the start and end of the line containing the
+    /// character, with any leading and trailing whitespace removed
+    pub trimmed_line: sg_utf8_bounds,
+}
+
+/// The offset of a character within a string (typically a line of source code), using several
+/// different units
+///
+/// All offsets are 0-indexed.
+#[repr(C)]
+#[derive(Clone, Copy, Default)]
+pub struct sg_offset {
+    /// The number of UTF-8-encoded bytes appearing before this character in the string
+    pub utf8_offset: usize,
+    /// The number of UTF-16 code units appearing before this character in the string
+    pub utf16_offset: usize,
+}
+
+/// A half-open range identifying a range of characters in a string.
+#[repr(C)]
+#[derive(Clone, Copy, Default)]
+pub struct sg_utf8_bounds {
+    /// The UTF-8 byte index of the first character in the range.
+    pub start: usize,
+    /// The UTF-8 byte index of the first character _after_ the range.
+    pub end: usize,
+}
+
+/// An array of all of the source information in a stack graph.  Source information is associated
+/// with nodes, so node handles are indices into this array.  It is _not_ guaranteed that there
+/// will an entry in this array for every node handle; if you have a node handle whose value is
+/// larger than `count`, then use a 0-valued `sg_source_info` if you need source information for
+/// that node.
+///
+/// There will never be a valid entry at index 0; a handle with the value 0 represents a missing
+/// node.
+#[repr(C)]
+pub struct sg_source_infos {
+    pub infos: *const sg_source_info,
+    pub count: usize,
+}
+
+/// Returns a reference to the array of source information in this stack graph.  The resulting
+/// array pointer is only valid until the next call to any function that mutates the stack graph.
+#[no_mangle]
+pub extern "C" fn sg_stack_graph_source_infos(graph: *const sg_stack_graph) -> sg_source_infos {
+    let graph = unsafe { &(*graph).inner };
+    sg_source_infos {
+        infos: graph.source_info.as_ptr() as *const sg_source_info,
+        count: graph.source_info.len(),
+    }
+}
+
+/// A tuple of a node handle and source information for that node.  Used with the
+/// `sg_add_source_info` function to add source information to a stack graph.
+#[repr(C)]
+#[derive(Clone, Copy)]
+pub struct sg_node_source_info {
+    pub node: sg_node_handle,
+    pub source_info: sg_source_info,
+}
+
+/// Adds new source information to the stack graph.  You provide an array of `sg_node_source_info`
+/// instances.  Any existing source information for any node mentioned in the array is overwritten.
+#[no_mangle]
+pub extern "C" fn sg_stack_graph_add_source_infos(
+    graph: *mut sg_stack_graph,
+    count: usize,
+    infos: *const sg_node_source_info,
+) {
+    let graph = unsafe { &mut (*graph).inner };
+    let infos = unsafe { std::slice::from_raw_parts(infos, count) };
+    for i in 0..count {
+        let node = unsafe { std::mem::transmute(infos[i].node) };
+        let info = graph.source_info_mut(node);
+        *info = unsafe { std::mem::transmute(infos[i].source_info) };
+    }
+}
+
 //-------------------------------------------------------------------------------------------------
 // Symbol stacks
 
@@ -559,7 +756,7 @@ pub struct sg_symbol_stack {
     /// The handle of the first element in the symbol stack, or SG_LIST_EMPTY_HANDLE if the list is
     /// empty, or 0 if the list is null.
     pub cells: sg_symbol_stack_cell_handle,
-    pub length: usize,
+    pub length: u32,
 }
 
 impl From<SymbolStack> for sg_symbol_stack {
@@ -644,6 +841,7 @@ pub struct sg_scope_stack {
     /// The handle of the first element in the scope stack, or SG_LIST_EMPTY_HANDLE if the list is
     /// empty, or 0 if the list is null.
     pub cells: sg_scope_stack_cell_handle,
+    pub length: u32,
 }
 
 impl From<ScopeStack> for sg_scope_stack {
@@ -744,7 +942,7 @@ pub struct sg_path_edge_list {
     /// empty, or 0 if the list is null.
     pub cells: sg_path_edge_list_cell_handle,
     pub direction: sg_deque_direction,
-    pub length: usize,
+    pub length: u32,
 }
 
 impl From<PathEdgeList> for sg_path_edge_list {
@@ -902,6 +1100,9 @@ pub extern "C" fn sg_path_arena_find_all_complete_paths(
 //-------------------------------------------------------------------------------------------------
 // Partial symbol stacks
 
+/// Represents an unknown list of scoped symbols.
+pub type sg_symbol_stack_variable = u32;
+
 /// A symbol with an unknown, but possibly empty, list of exported scopes attached to it.
 #[repr(C)]
 #[derive(Clone, Copy, Eq, PartialEq)]
@@ -931,6 +1132,12 @@ pub struct sg_partial_symbol_stack {
     /// list is empty, or 0 if the list is null.
     pub cells: sg_partial_symbol_stack_cell_handle,
     pub direction: sg_deque_direction,
+    pub length: u32,
+    /// The symbol stack variable representing the unknown content of a partial symbol stack, or 0
+    /// if the variable is missing.  (If so, this partial symbol stack can only match a symbol
+    /// stack with exactly the list of symbols in `cells`, instead of any symbol stack with those
+    /// symbols as a prefix.)
+    pub variable: sg_symbol_stack_variable,
 }
 
 impl From<PartialSymbolStack> for sg_partial_symbol_stack {
@@ -981,7 +1188,8 @@ pub extern "C" fn sg_partial_path_arena_partial_symbol_stack_cells(
 /// arrays.  The `lengths` array must have `count` elements, and provides the number of symbols in
 /// each partial symbol stack.  The `symbols` array contains the contents of each of these partial
 /// symbol stacks in one contiguous array.  Its length must be the sum of all of the counts in the
-/// `lengths` array.
+/// `lengths` array.  The `variables` array must have `count` elements, and provides the optional
+/// symbol stack variable for each partial symbol stack.
 ///
 /// You must also provide an `out` array, which must also have room for `count` elements.  We will
 /// fill this array in with the `sg_partial_symbol_stack` instances for each partial symbol stack
@@ -992,15 +1200,21 @@ pub extern "C" fn sg_partial_path_arena_add_partial_symbol_stacks(
     count: usize,
     mut symbols: *const sg_partial_scoped_symbol,
     lengths: *const usize,
+    variables: *const sg_symbol_stack_variable,
     out: *mut sg_partial_symbol_stack,
 ) {
     let partials = unsafe { &mut (*partials).inner };
     let lengths = unsafe { std::slice::from_raw_parts(lengths, count) };
+    let variables = unsafe { std::slice::from_raw_parts(variables, count) };
     let out = unsafe { std::slice::from_raw_parts_mut(out, count) };
     for i in 0..count {
         let length = lengths[i];
         let symbols_slice = unsafe { std::slice::from_raw_parts(symbols, length) };
-        let mut stack = PartialSymbolStack::empty();
+        let mut stack = if variables[i] == 0 {
+            PartialSymbolStack::empty()
+        } else {
+            PartialSymbolStack::from_variable(variables[i].try_into().unwrap())
+        };
         for j in 0..length {
             let symbol = symbols_slice[j].into();
             stack.push_back(partials, symbol);
@@ -1028,6 +1242,7 @@ pub struct sg_partial_scope_stack {
     /// list is empty, or 0 if the list is null.
     pub cells: sg_partial_scope_stack_cell_handle,
     pub direction: sg_deque_direction,
+    pub length: u32,
     /// The scope stack variable representing the unknown content of a partial scope stack, or 0 if
     /// the variable is missing.  (If so, this partial scope stack can only match a scope stack
     /// with exactly the list of scopes in `cells`, instead of any scope stack with those scopes as
@@ -1148,7 +1363,7 @@ pub struct sg_partial_path_edge_list {
     /// empty, or 0 if the list is null.
     pub cells: sg_partial_path_edge_list_cell_handle,
     pub direction: sg_deque_direction,
-    pub length: usize,
+    pub length: u32,
 }
 
 impl From<PartialPathEdgeList> for sg_partial_path_edge_list {
@@ -1336,6 +1551,29 @@ pub extern "C" fn sg_partial_path_arena_find_partial_paths_in_file(
 /// partial path.
 pub type sg_partial_path_handle = u32;
 
+/// An array of all of the partial paths in a partial path database.  Partial path handles are
+/// indices into this array.  There will never be a valid partial path at index 0; a handle with
+/// the value 0 represents a missing partial path.
+#[repr(C)]
+pub struct sg_partial_paths {
+    pub paths: *const sg_partial_path,
+    pub count: usize,
+}
+
+/// Returns a reference to the array of partial path data in this partial path database.  The
+/// resulting array pointer is only valid until the next call to any function that mutates the
+/// partial path database.
+#[no_mangle]
+pub extern "C" fn sg_partial_path_database_partial_paths(
+    db: *const sg_partial_path_database,
+) -> sg_partial_paths {
+    let db = unsafe { &(*db).inner };
+    sg_partial_paths {
+        paths: db.partial_paths.as_ptr() as *const sg_partial_path,
+        count: db.partial_paths.len(),
+    }
+}
+
 /// Adds new partial paths to the partial path database.  `paths` is the array of partial paths
 /// that you want to add; `count` is the number of them.
 ///
@@ -1345,6 +1583,10 @@ pub type sg_partial_path_handle = u32;
 /// You should take care not to add a partial path to the database multiple times.  This won't
 /// cause an _error_, in that nothing will break, but it will probably cause you to get duplicate
 /// paths from the path-stitching algorithm.
+///
+/// You must also provide an `out` array, which must also have room for `count` elements.  We will
+/// fill this array in with the `sg_partial_path_edge_list` instances for each partial path edge
+/// list that is created.
 #[no_mangle]
 pub extern "C" fn sg_partial_path_database_add_partial_paths(
     graph: *const sg_stack_graph,
@@ -1352,13 +1594,85 @@ pub extern "C" fn sg_partial_path_database_add_partial_paths(
     db: *mut sg_partial_path_database,
     count: usize,
     paths: *const sg_partial_path,
+    out: *mut sg_partial_path_handle,
 ) {
     let graph = unsafe { &(*graph).inner };
     let partials = unsafe { &mut (*partials).inner };
     let db = unsafe { &mut (*db).inner };
     let paths = unsafe { std::slice::from_raw_parts(paths, count) };
+    let out = unsafe { std::slice::from_raw_parts_mut(out as *mut Handle<PartialPath>, count) };
     for i in 0..count {
-        db.add_partial_path(graph, partials, paths[i].into());
+        out[i] = db.add_partial_path(graph, partials, paths[i].into());
+    }
+}
+
+//-------------------------------------------------------------------------------------------------
+// Local nodes
+
+/// Encodes a set of node handles.
+///
+/// The elements are encoded in a bit set.  Use the traditional mask and shift pattern to determine
+/// if a particular handle is in the set:
+///
+/// ``` c
+/// size_t element_index = handle / 32;
+/// size_t bit_index = handle % 32;
+/// size_t bit_mask = 1 << bit_index;
+/// bool bit_is_set =
+///     element_index < set.length &&
+///     (set.elements[element_index] & bit_mask) != 0;
+/// ```
+#[repr(C)]
+pub struct sg_node_handle_set {
+    pub elements: *const u32,
+    /// Note that this is the number of uint32_t's in `elements`, NOT the number of bits in the set.
+    pub length: usize,
+}
+
+/// Determines which nodes in the stack graph are “local”, taking into account the partial paths in
+/// this database.  The result is valid until the next call to this function, or until the database
+/// is freed.
+///
+/// A local node has no partial path that connects it to the root node in either direction. That
+/// means that it cannot participate in any paths that leave the file.
+///
+/// This method is meant to be used at index time, to calculate the set of nodes that are local
+/// after having just calculated the set of partial paths for the file.
+#[no_mangle]
+pub extern "C" fn sg_partial_path_database_find_local_nodes(db: *mut sg_partial_path_database) {
+    let db = unsafe { &mut (*db).inner };
+    db.find_local_nodes();
+}
+
+/// Marks that a list of stack graph nodes are local.
+///
+/// This method is meant to be used at query time.  You will have precalculated the set of local
+/// nodes for a file at index time; at query time, you will load this information from your storage
+/// layer and use this method to update our internal view of which nodes are local.
+#[no_mangle]
+pub extern "C" fn sg_partial_path_database_mark_local_nodes(
+    db: *mut sg_partial_path_database,
+    count: usize,
+    nodes: *const sg_node_handle,
+) {
+    let db = unsafe { &mut (*db).inner };
+    let nodes = unsafe { std::slice::from_raw_parts(nodes, count) };
+    for node in nodes {
+        db.mark_local_node(node.clone().into());
+    }
+}
+
+/// Returns a reference to the set of stack graph nodes that are local, according to this database
+/// of partial paths.  The resulting set is only valid until the next call to any function that
+/// mutates the partial path database.
+#[no_mangle]
+pub extern "C" fn sg_partial_path_database_local_nodes(
+    db: *const sg_partial_path_database,
+) -> sg_node_handle_set {
+    let db = unsafe { &(*db).inner };
+    sg_node_handle_set {
+        elements: db.local_nodes.as_ptr(),
+        length: db.local_nodes.len(),
     }
 }
 
@@ -1408,17 +1722,22 @@ struct ForwardPathStitcher {
 }
 
 impl ForwardPathStitcher {
-    fn new(stitcher: PathStitcher) -> ForwardPathStitcher {
+    fn new(stitcher: PathStitcher, paths: &mut Paths) -> ForwardPathStitcher {
         let mut this = ForwardPathStitcher {
             previous_phase_paths: std::ptr::null(),
             previous_phase_paths_length: 0,
             stitcher,
         };
-        this.update_previous_phase_paths();
+        this.update_previous_phase_paths(paths);
         this
     }
 
-    fn update_previous_phase_paths(&mut self) {
+    fn update_previous_phase_paths(&mut self, paths: &mut Paths) {
+        for path in self.stitcher.previous_phase_paths_slice_mut() {
+            // Each path's edge list comes out in backwards order.  Requesting a forwards iterator
+            // before we return ensures that it will also be available in forwards order.
+            let _ = path.edges.iter(paths);
+        }
         let slice = self.stitcher.previous_phase_paths_slice();
         self.previous_phase_paths = slice.as_ptr();
         self.previous_phase_paths_length = slice.len();
@@ -1455,7 +1774,7 @@ pub extern "C" fn sg_forward_path_stitcher_new(
         db,
         starting_nodes.iter().copied().map(sg_node_handle::into),
     );
-    Box::into_raw(Box::new(ForwardPathStitcher::new(stitcher))) as *mut _
+    Box::into_raw(Box::new(ForwardPathStitcher::new(stitcher, paths))) as *mut _
 }
 
 /// Runs the next phase of the path-stitching algorithm.  We will have built up a set of
@@ -1482,7 +1801,7 @@ pub extern "C" fn sg_forward_path_stitcher_process_next_phase(
     stitcher
         .stitcher
         .process_next_phase(graph, paths, partials, db);
-    stitcher.update_previous_phase_paths();
+    stitcher.update_previous_phase_paths(paths);
 }
 
 /// Frees a forward path stitcher.
@@ -1490,3 +1809,136 @@ pub extern "C" fn sg_forward_path_stitcher_process_next_phase(
 pub extern "C" fn sg_forward_path_stitcher_free(stitcher: *mut sg_forward_path_stitcher) {
     drop(unsafe { Box::from_raw(stitcher as *mut ForwardPathStitcher) });
 }
+
+//-------------------------------------------------------------------------------------------------
+// Forward partial path stitching
+
+/// Implements a phased forward partial path stitching algorithm.
+///
+/// Our overall goal is to start with a set of _seed_ partial paths, and to repeatedly extend each
+/// partial path by concatenating another, compatible partial path onto the end of it.  (If there
+/// are multiple compatible partial paths, we concatenate each of them separately, resulting in
+/// more than one extension for the current path.)
+///
+/// We perform this processing in _phases_.  At the start of each phase, we have a _current set_ of
+/// partial paths that need to be processed.  As we extend those partial paths, we add the
+/// extensions to the set of partial paths to process in the _next_ phase.  Phases are processed
+/// one at a time, each time you invoke `sg_forward_partial_path_stitcher_process_next_phase`.
+///
+/// After each phase has completed, the `previous_phase_paths` and `previous_phase_paths_length`
+/// fields give you all of the partial paths that were discovered during that phase.  That gives
+/// you a chance to add to the `sg_partial_path_database` all of the other partial paths that we
+/// might need to extend those partial paths with before invoking the next phase.
+#[repr(C)]
+pub struct sg_forward_partial_path_stitcher {
+    /// The new candidate partial paths that were discovered in the most recent phase.
+    pub previous_phase_partial_paths: *const sg_partial_path,
+    /// The number of new candidate partial paths that were discovered in the most recent phase.
+    /// If this is 0, then the partial path stitching algorithm is complete.
+    pub previous_phase_partial_paths_length: usize,
+}
+
+// This is the Rust equivalent of a common C trick, where you have two versions of a struct — a
+// publicly visible one and a private one containing internal implementation details.  In our case,
+// `sg_forward_partial_path_stitcher` is the public struct, and
+// `InternalForwardPartialPathStitcher` is the internal one.  The main requirement is that the
+// private struct must start with a copy of all of the fields in the public struct — ensuring that
+// those fields occur at the same offset in both.  The private struct can contain additional
+// (private) fields, but they must appear _after_ all of the publicly visible fields.
+//
+// In our case, we do this because we don't want to expose the existence or details of the
+// ForwardPartialPathStitcher type via the C API.
+#[repr(C)]
+struct InternalForwardPartialPathStitcher {
+    previous_phase_partial_paths: *const PartialPath,
+    previous_phase_partial_paths_length: usize,
+    stitcher: ForwardPartialPathStitcher,
+}
+
+impl InternalForwardPartialPathStitcher {
+    fn new(
+        stitcher: ForwardPartialPathStitcher,
+        partials: &mut PartialPaths,
+    ) -> InternalForwardPartialPathStitcher {
+        let mut this = InternalForwardPartialPathStitcher {
+            previous_phase_partial_paths: std::ptr::null(),
+            previous_phase_partial_paths_length: 0,
+            stitcher,
+        };
+        this.update_previous_phase_partial_paths(partials);
+        this
+    }
+
+    fn update_previous_phase_partial_paths(&mut self, partials: &mut PartialPaths) {
+        for path in self.stitcher.previous_phase_partial_paths_slice_mut() {
+            path.ensure_both_directions(partials);
+        }
+        let slice = self.stitcher.previous_phase_partial_paths_slice();
+        self.previous_phase_partial_paths = slice.as_ptr();
+        self.previous_phase_partial_paths_length = slice.len();
+    }
+}
+
+/// Creates a new forward partial path stitcher that is "seeded" with a set of starting stack graph
+/// nodes.
+///
+/// Before calling this method, you must ensure that `db` contains all of the possible partial
+/// paths that start with any of your requested starting nodes.
+///
+/// Before calling `sg_forward_partial_path_stitcher_process_next_phase` for the first time, you
+/// must ensure that `db` contains all possible extensions of any of those initial partial paths.
+/// You can retrieve a list of those extensions via the `previous_phase_partial_paths` and
+/// `previous_phase_partial_paths_length` fields.
+#[no_mangle]
+pub extern "C" fn sg_forward_partial_path_stitcher_new(
+    graph: *const sg_stack_graph,
+    partials: *mut sg_partial_path_arena,
+    db: *mut sg_partial_path_database,
+    count: usize,
+    starting_nodes: *const sg_node_handle,
+) -> *mut sg_forward_partial_path_stitcher {
+    let graph = unsafe { &(*graph).inner };
+    let partials = unsafe { &mut (*partials).inner };
+    let db = unsafe { &mut (*db).inner };
+    let starting_nodes = unsafe { std::slice::from_raw_parts(starting_nodes, count) };
+    let stitcher = ForwardPartialPathStitcher::new(
+        graph,
+        partials,
+        db,
+        starting_nodes.iter().copied().map(sg_node_handle::into),
+    );
+    Box::into_raw(Box::new(InternalForwardPartialPathStitcher::new(
+        stitcher, partials,
+    ))) as *mut _
+}
+
+/// Runs the next phase of the algorithm.  We will have built up a set of incomplete partial paths
+/// during the _previous_ phase.  Before calling this function, you must ensure that `db` contains
+/// all of the possible partial paths that we might want to extend any of those candidate partial
+/// paths with.
+///
+/// After this method returns, you can retrieve a list of the (possibly incomplete) partial paths
+/// that were encountered during this phase via the `previous_phase_partial_paths` and
+/// `previous_phase_partial_paths_length` fields.
+#[no_mangle]
+pub extern "C" fn sg_forward_partial_path_stitcher_process_next_phase(
+    graph: *const sg_stack_graph,
+    partials: *mut sg_partial_path_arena,
+    db: *mut sg_partial_path_database,
+    stitcher: *mut sg_forward_partial_path_stitcher,
+) {
+    let graph = unsafe { &(*graph).inner };
+    let partials = unsafe { &mut (*partials).inner };
+    let db = unsafe { &mut (*db).inner };
+    let stitcher = unsafe { &mut *(stitcher as *mut InternalForwardPartialPathStitcher) };
+    stitcher.stitcher.process_next_phase(graph, partials, db);
+    stitcher.update_previous_phase_partial_paths(partials);
+}
+
+/// Frees a forward path stitcher.
+#[no_mangle]
+pub extern "C" fn sg_forward_partial_path_stitcher_free(
+    stitcher: *mut sg_forward_partial_path_stitcher,
+) {
+    drop(unsafe { Box::from_raw(stitcher as *mut InternalForwardPartialPathStitcher) });
+}
