@@ -18,6 +18,9 @@
 use std::collections::VecDeque;
 use std::fmt::Display;
 
+use controlled_option::ControlledOption;
+use controlled_option::Niche;
+
 use crate::arena::Deque;
 use crate::arena::DequeArena;
 use crate::arena::Handle;
@@ -126,13 +129,17 @@ where
 #[derive(Clone, Copy)]
 pub struct ScopedSymbol {
     pub symbol: Handle<Symbol>,
-    pub scopes: Option<ScopeStack>,
+    pub scopes: ControlledOption<ScopeStack>,
 }
 
 impl ScopedSymbol {
     pub fn equals(&self, paths: &Paths, other: &ScopedSymbol) -> bool {
         self.symbol == other.symbol
-            && equals_option(self.scopes, other.scopes, |a, b| a.equals(paths, &b))
+            && equals_option(
+                self.scopes.into_option(),
+                other.scopes.into_option(),
+                |a, b| a.equals(paths, &b),
+            )
     }
 
     pub fn cmp(
@@ -143,7 +150,13 @@ impl ScopedSymbol {
     ) -> std::cmp::Ordering {
         std::cmp::Ordering::Equal
             .then_with(|| graph[self.symbol].cmp(&graph[other.symbol]))
-            .then_with(|| cmp_option(self.scopes, other.scopes, |a, b| a.cmp(paths, &b)))
+            .then_with(|| {
+                cmp_option(
+                    self.scopes.into_option(),
+                    other.scopes.into_option(),
+                    |a, b| a.cmp(paths, &b),
+                )
+            })
     }
 
     pub fn display<'a>(self, graph: &'a StackGraph, paths: &'a mut Paths) -> impl Display + 'a {
@@ -153,8 +166,9 @@ impl ScopedSymbol {
 
 impl DisplayWithPaths for ScopedSymbol {
     fn prepare(&mut self, graph: &StackGraph, paths: &mut Paths) {
-        if let Some(scopes) = &mut self.scopes {
+        if let Some(mut scopes) = self.scopes.into_option() {
             scopes.prepare(graph, paths);
+            self.scopes = scopes.into();
         }
     }
 
@@ -164,10 +178,10 @@ impl DisplayWithPaths for ScopedSymbol {
         paths: &Paths,
         f: &mut std::fmt::Formatter,
     ) -> std::fmt::Result {
-        match self.scopes {
+        match self.scopes.into_option() {
             Some(scopes) => write!(
                 f,
-                "{}/{}",
+                "{}/({})",
                 self.symbol.display(graph),
                 display_prepared(scopes, graph, paths),
             ),
@@ -179,10 +193,11 @@ impl DisplayWithPaths for ScopedSymbol {
 /// A sequence of symbols that describe what we are currently looking for while in the middle of
 /// the path-finding algorithm.
 #[repr(C)]
-#[derive(Clone, Copy)]
+#[derive(Clone, Copy, Niche)]
 pub struct SymbolStack {
+    #[niche]
     list: List<ScopedSymbol>,
-    length: usize,
+    length: u32,
 }
 
 impl SymbolStack {
@@ -194,7 +209,7 @@ impl SymbolStack {
 
     #[inline(always)]
     pub fn len(&self) -> usize {
-        self.length
+        self.length as usize
     }
 
     /// Returns an empty symbol stack.
@@ -276,9 +291,11 @@ impl DisplayWithPaths for SymbolStack {
 
 /// A sequence of exported scopes, used to pass name-binding context around a stack graph.
 #[repr(C)]
-#[derive(Clone, Copy)]
+#[derive(Clone, Copy, Niche)]
 pub struct ScopeStack {
+    #[niche]
     list: List<Handle<Node>>,
+    length: u32,
 }
 
 impl ScopeStack {
@@ -288,10 +305,16 @@ impl ScopeStack {
         self.list.is_empty()
     }
 
+    #[inline(always)]
+    pub fn len(&self) -> usize {
+        self.length as usize
+    }
+
     /// Returns an empty scope stack.
     pub fn empty() -> ScopeStack {
         ScopeStack {
             list: List::empty(),
+            length: 0,
         }
     }
 
@@ -310,13 +333,18 @@ impl ScopeStack {
     ///
     /// [`Node`]: ../graph/enum.Node.html
     pub fn push_front(&mut self, paths: &mut Paths, node: Handle<Node>) {
+        self.length += 1;
         self.list.push_front(&mut paths.scope_stacks, node);
     }
 
     /// Removes and returns the [`Node`][] at the front of this scope stack.  If the stack is
     /// empty, returns `None`.
     pub fn pop_front(&mut self, paths: &Paths) -> Option<Handle<Node>> {
-        self.list.pop_front(&paths.scope_stacks).copied()
+        let result = self.list.pop_front(&paths.scope_stacks).copied();
+        if result.is_some() {
+            self.length -= 1;
+        }
+        result
     }
 
     pub fn display<'a>(self, graph: &'a StackGraph, paths: &'a mut Paths) -> impl Display + 'a {
@@ -386,10 +414,11 @@ impl DisplayWithPaths for PathEdge {
 /// The edges in a path keep track of precedence information so that we can correctly handle
 /// shadowed definitions.
 #[repr(C)]
-#[derive(Clone, Copy)]
+#[derive(Clone, Copy, Niche)]
 pub struct PathEdgeList {
+    #[niche]
     edges: Deque<PathEdge>,
-    length: usize,
+    length: u32,
 }
 
 impl PathEdgeList {
@@ -401,7 +430,13 @@ impl PathEdgeList {
 
     #[inline(always)]
     pub fn len(&self) -> usize {
-        self.length
+        self.length as usize
+    }
+
+    /// Returns whether this edge list is iterable in both directions without needing mutable
+    /// access to the arena.
+    pub fn have_reversal(&self, paths: &Paths) -> bool {
+        self.edges.have_reversal(&paths.path_edges)
     }
 
     /// Returns an empty edge list.
@@ -558,7 +593,7 @@ impl Path {
                     paths,
                     ScopedSymbol {
                         symbol: node.symbol,
-                        scopes: Some(scope_stack),
+                        scopes: ControlledOption::some(scope_stack),
                     },
                 );
             }
@@ -567,7 +602,7 @@ impl Path {
                     paths,
                     ScopedSymbol {
                         symbol: node.symbol,
-                        scopes: None,
+                        scopes: ControlledOption::none(),
                     },
                 );
             }
@@ -718,7 +753,7 @@ impl Path {
             let sink_symbol = sink.symbol;
             let scoped_symbol = ScopedSymbol {
                 symbol: sink_symbol,
-                scopes: None,
+                scopes: ControlledOption::none(),
             };
             self.symbol_stack.push_front(paths, scoped_symbol);
         } else if let Node::PushScopedSymbol(sink) = sink {
@@ -730,7 +765,7 @@ impl Path {
             attached_scopes.push_front(paths, sink_scope);
             let scoped_symbol = ScopedSymbol {
                 symbol: sink_symbol,
-                scopes: Some(attached_scopes),
+                scopes: ControlledOption::some(attached_scopes),
             };
             self.symbol_stack.push_front(paths, scoped_symbol);
         } else if let Node::PopSymbol(sink) = sink {
@@ -752,7 +787,7 @@ impl Path {
             if top.symbol != sink.symbol {
                 return Err(PathResolutionError::IncorrectPoppedSymbol);
             }
-            let new_scope_stack = match top.scopes {
+            let new_scope_stack = match top.scopes.into_option() {
                 Some(scopes) => scopes,
                 None => return Err(PathResolutionError::MissingAttachedScopeList),
             };
